# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import logging
import shutil
from typing import Dict, cast

from rebasehelper.logger import CustomLogger
from rebasehelper.helpers.path_helper import PathHelper


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


class TemporaryEnvironment:

    """Class representing a temporary environment (directory) that can be used as a workspace.

    Works as a context manager.

    """

    TEMPDIR: str = 'TEMPDIR'

    def __init__(self, exit_callback=None):
        self._env: Dict[str, str] = {}
        self._exit_callback = exit_callback

    def __enter__(self):
        self._env[self.TEMPDIR] = PathHelper.get_temp_dir()
        logger.debug("Created environment in '%s'", self.path())
        return self

    def __exit__(self, *args):
        # run callback before removing the environment
        try:
            self._exit_callback(**self.env())
        except TypeError:
            pass
        else:
            logger.debug("Exit callback executed successfully")

        shutil.rmtree(self.path(), onerror=lambda func, path, excinfo: shutil.rmtree(path))
        logger.debug("Destroyed environment in '%s'", self.path())

    def __str__(self):
        return "<TemporaryEnvironment path='{}'>".format(self.path())

    def path(self):
        """Gets the path to the temporary environment.

        Returns:
            str: Absolute path to the temporary environment.

        """
        return self._env.get(self.TEMPDIR, '')

    def env(self):
        """Gets a copy of _env dictionary.

        Returns:
            dict: Copy of _env dictionary.

        """
        return self._env.copy()



#!/usr/bin/env python
# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import distutils.command.sdist
import os
import types

import pkg_resources
from setuptools import setup, find_packages


def get_rpm_distribution():
    for distribution in ['rpm', 'rpm-python']:
        try:
            pkg_resources.get_distribution(distribution)
        except pkg_resources.DistributionNotFound:
            continue
        else:
            return distribution
    return 'rpm-py-installer'


def get_requirements():
    result = [
        'ansicolors',
        'copr',
        'pyquery',
        'python-pam',
        'requests',
        'GitPython',
    ]
    # there is no rpm nor gssapi inside RTD build environment
    if not os.getenv('READTHEDOCS'):
        result.append(get_rpm_distribution())
        result.append('requests-gssapi')
    return result


def get_readme():
    with open('README.md') as f:
        return f.read()


class SDistCommand(distutils.command.sdist.sdist):

    user_options = distutils.command.sdist.sdist.user_options + \
        [('base-name=', 'b', 'override base name of the distribution')]

    def initialize_options(self):
        super(SDistCommand, self).initialize_options()
        self.base_name = None

    def make_distribution(self):
        if not self.base_name:
            super(SDistCommand, self).make_distribution()
            return
        metadata = self.distribution.metadata
        # temporarily patch get_name() method
        get_name = metadata.get_name
        metadata.get_name = types.MethodType(lambda _: self.base_name, metadata)
        try:
            super(SDistCommand, self).make_distribution()
        finally:
            metadata.get_name = get_name


setup(
    name='rebasehelper',
    description='This tool helps you rebase your package to the latest version',
    long_description=get_readme(),
    long_description_content_type='text/markdown',
    keywords=['rebase', 'packages', 'rpm'],
    author='Petr Hracek',
    author_email='phracek@redhat.com',
    url='https://github.com/rebase-helper/rebase-helper',
    license='GPLv2+',
    packages=find_packages(),
    include_package_data=True,
    use_scm_version=True,
    cmdclass={
        'sdist': SDistCommand,
    },
    entry_points={
        'console_scripts': [
            'rebase-helper = rebasehelper.cli:CliHelper.run',
        ],
        'rebasehelper.build_tools': [
            'rpmbuild = rebasehelper.plugins.build_tools.rpm.rpmbuild:Rpmbuild',
            'mock = rebasehelper.plugins.build_tools.rpm.mock:Mock',
            'koji = rebasehelper.plugins.build_tools.rpm.koji_:Koji',
            'copr = rebasehelper.plugins.build_tools.rpm.copr_:Copr',
        ],
        'rebasehelper.srpm_build_tools': [
            'rpmbuild = rebasehelper.plugins.build_tools.srpm.rpmbuild:Rpmbuild',
            'mock = rebasehelper.plugins.build_tools.srpm.mock:Mock',
        ],
        'rebasehelper.checkers': [
            'rpmdiff = rebasehelper.plugins.checkers.rpmdiff:RpmDiff',
            'pkgdiff = rebasehelper.plugins.checkers.pkgdiff:PkgDiff',
            'abipkgdiff = rebasehelper.plugins.checkers.abipkgdiff:AbiPkgDiff',
            'csmock = rebasehelper.plugins.checkers.csmock:CsMock',
            'licensecheck = rebasehelper.plugins.checkers.licensecheck:LicenseCheck',
            'sonamecheck = rebasehelper.plugins.checkers.sonamecheck:SonameCheck',
            'rpminspect-rpm = rebasehelper.plugins.checkers.rpminspect_rpm:RpminspectRpm',
            'rpminspect-srpm = rebasehelper.plugins.checkers.rpminspect_srpm:RpminspectSrpm',
        ],
        'rebasehelper.spec_hooks': [
            'typo-fix = rebasehelper.plugins.spec_hooks.typo_fix:TypoFix',
            'pypi-url-fix = rebasehelper.plugins.spec_hooks.pypi_url_fix:PyPIURLFix',
            'ruby-helper = rebasehelper.plugins.spec_hooks.ruby_helper:RubyHelper',
            'commit-hash-updater = rebasehelper.plugins.spec_hooks.commit_hash_updater:CommitHashUpdater',
            'paths-to-rpm-macros = rebasehelper.plugins.spec_hooks.paths_to_rpm_macros:PathsToRPMMacros',
            'escape-macros = rebasehelper.plugins.spec_hooks.escape_macros:EscapeMacros',
            'replace-old-version = rebasehelper.plugins.spec_hooks.replace_old_version:ReplaceOldVersion',
        ],
        'rebasehelper.build_log_hooks': [
            'files = rebasehelper.plugins.build_log_hooks.files:Files',
        ],
        'rebasehelper.versioneers': [
            'anitya = rebasehelper.plugins.versioneers.anitya:Anitya',
            'pypi = rebasehelper.plugins.versioneers.pypi:PyPI',
            'rubygems = rebasehelper.plugins.versioneers.rubygems:RubyGems',
            'npmjs = rebasehelper.plugins.versioneers.npmjs:NPMJS',
            'cpan = rebasehelper.plugins.versioneers.cpan:CPAN',
            'hackage = rebasehelper.plugins.versioneers.hackage:Hackage',
        ],
        'rebasehelper.output_tools': [
            'json = rebasehelper.plugins.output_tools.json_:JSON',
            'text = rebasehelper.plugins.output_tools.text:Text',
        ]
    },
    install_requires=get_requirements(),
    setup_requires=[
        'setuptools_scm',
        'setuptools_scm_git_archive',
    ],
    classifiers=[
        'Development Status :: 4 - Beta',
        'Environment :: Console',
        'Intended Audience :: Developers',
        'License :: OSI Approved :: GNU General Public License v2 or later (GPLv2+)',
        'Operating System :: POSIX :: Linux',
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Topic :: Software Development',
        'Topic :: System :: Operating System',
        'Topic :: System :: Software Distribution',
        'Topic :: Utilities',
    ],
    project_urls={
        'Source Code': 'https://github.com/rebase-helper/rebase-helper',
        'Documentation': 'https://rebase-helper.readthedocs.io',
        'Bug Tracker': 'https://github.com/rebase-helper/rebase-helper/issues',
    }
)



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import argparse
import collections
import enum
import itertools
import logging
import os
import re
import shlex
import shutil
import urllib.parse
from datetime import date
from difflib import SequenceMatcher
from operator import itemgetter
from typing import List, Optional, Pattern, Tuple, Dict, Union, cast

import rpm  # type: ignore

from rebasehelper import constants
from rebasehelper.archive import Archive
from rebasehelper.spec_content import SpecContent
from rebasehelper.tags import Tag, Tags
from rebasehelper.exceptions import RebaseHelperError, DownloadError, ParseError, LookasideCacheError
from rebasehelper.argument_parser import SilentArgumentParser
from rebasehelper.logger import CustomLogger
from rebasehelper.helpers.download_helper import DownloadHelper
from rebasehelper.helpers.macro_helper import MacroHelper
from rebasehelper.helpers.rpm_helper import RpmHelper, RpmHeader
from rebasehelper.helpers.git_helper import GitHelper
from rebasehelper.helpers.lookaside_cache_helper import LookasideCacheHelper


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


def get_rebase_name(dir_name, name):
    """
    Function returns a name in results directory

    :param dir_name:
    :param name:
    :return: full path to results dir with name
    """
    file_name = os.path.basename(name)
    return os.path.join(dir_name, file_name)


class PatchList(list):
    def _get_index_list(self, item):
        for x in self:
            if x.index == item.index:
                return x

    def __getitem__(self, item):
        return super(PatchList, self).__getitem__(self._get_index_list(item))


class PatchObject:

    """Class represents set of information about patches"""

    def __init__(self, path, index, strip):
        self.path = path
        self.index = index
        self.strip = strip

    def get_patch_name(self):
        return os.path.basename(self.path)


class PackageCategory(enum.Enum):
    python: Pattern[str] = re.compile(r'^python[23]?-')
    perl: Pattern[str] = re.compile(r'^perl-')
    ruby: Pattern[str] = re.compile(r'^rubygem-')
    nodejs: Pattern[str] = re.compile(r'^nodejs-')
    php: Pattern[str] = re.compile(r'^php-')
    haskell: Pattern[str] = re.compile(r'^ghc-')
    R: Pattern[str] = re.compile(r'^R-')
    rust: Pattern[str] = re.compile(r'^rust-')


def saves(func):
    """Decorator for saving the SpecFile after a method is run."""
    def wrapper(spec, *args, **kwargs):
        func(spec, *args, **kwargs)
        spec.save()

    return wrapper


class SpecFile:

    """Class representing a SPEC file. Be aware that using SpecFile
    modifies RPM macros in global context."""

    def __init__(self, path: str, sources_location: str = '', predefined_macros: Optional[Dict[str, str]] = None):
        # Initialize attributes
        self.path: str = path
        self.sources_location: str = sources_location
        self.predefined_macros: Dict[str, str] = predefined_macros or {}
        self.prep_section: str = ''
        self.sources: List[str] = []
        self.patches: Dict[str, List[PatchObject]] = {}
        self.removed_patches: List[str] = []
        self.category: Optional[PackageCategory] = None
        self.spc: rpm.spec = RpmHelper.get_rpm_spec(self.path, self.sources_location, self.predefined_macros)
        self.header: RpmHeader = RpmHeader(self.spc.sourceHeader)
        self.spec_content: SpecContent = self._read_spec_content()
        self.tags: Tags = Tags(self.spec_content, SpecContent(self.spc.parsed))

        # Load rpm information
        self._update_data()

    def __del__(self):
        # make sure there are no leftover macros
        rpm.reloadConfig()

    def download_remote_sources(self):
        """
        Method that iterates over all sources and downloads ones, which contain URL instead of just a file.

        :return: None
        """
        try:
            # try to download old sources from Fedora lookaside cache
            LookasideCacheHelper.download('fedpkg', os.path.dirname(self.path), self.header.name, self.sources_location)
        except LookasideCacheError as e:
            logger.verbose("Downloading sources from lookaside cache failed. "
                           "Reason: %s.", str(e))

        # filter out only sources with URL
        remote_files = [source for source in self.sources if bool(urllib.parse.urlparse(source).scheme)]
        # download any sources that are not yet downloaded
        for remote_file in remote_files:
            local_file = os.path.join(self.sources_location, os.path.basename(remote_file))
            if not os.path.isfile(local_file):
                logger.verbose("File '%s' doesn't exist locally, downloading it.", local_file)
                try:
                    DownloadHelper.download_file(remote_file, local_file)
                except DownloadError as e:
                    raise RebaseHelperError("Failed to download file from URL {}. "
                                            "Reason: '{}'. ".format(remote_file, str(e))) from e

    def update(self) -> None:
        # explicitly discard old instance to prevent rpm from destroying
        # "sources" and "patches" lua tables after new instance is created
        self.spc = None
        self.spc = RpmHelper.get_rpm_spec(self.path, self.sources_location, self.predefined_macros)
        self.header = RpmHeader(self.spc.sourceHeader)
        self.spec_content = self._read_spec_content()
        self.tags = Tags(self.spec_content, SpecContent(self.spc.parsed))
        self._update_data()

    def _update_data(self):
        """
        Function updates data from given SPEC file

        :return:
        """
        def guess_category():
            for pkg in self.spc.packages:
                header = RpmHeader(pkg.header)
                for category in PackageCategory:
                    if category.value.match(header.name):
                        return category
                    for provide in header.providename:
                        if category.value.match(provide):
                            return category
            return None
        self.category = guess_category()
        self.sources = self._get_spec_sources_list(self.spc)
        self.prep_section = self.spc.prep
        self.main_source_index = self._identify_main_source(self.spc)
        self.patches = self._get_initial_patches()
        self.macros = MacroHelper.dump()

    ######################
    # TAG HELPER METHODS #
    ######################

    def tag(self, name: str, section: Optional[Union[str, int]] = None) -> Optional[Tag]:
        """Returns the first non-unique tag."""
        if isinstance(section, str):
            tags = self.tags.filter(section_name=section, name=name)
        else:
            tags = self.tags.filter(section_index=section, name=name)
        return next(tags, None)

    def get_raw_tag_value(self, tag_name: str, section: Optional[Union[str, int]] = None) -> Optional[str]:
        tag = self.tag(tag_name, section)
        if not tag:
            return None
        return self.spec_content[tag.section_index][tag.line][slice(*tag.value_span)]

    def set_raw_tag_value(self, tag_name: str, value: str, section: Optional[Union[str, int]] = None) -> None:
        tag = self.tag(tag_name, section)
        if not tag:
            return
        sec = self.spec_content[tag.section_index]
        line = sec[tag.line]
        sec[tag.line] = line[:tag.value_span[0]] + value + line[tag.value_span[1]:]
        # update span
        tag.value_span = (tag.value_span[0], tag.value_span[0] + len(value))

    ###########################
    # SOURCES RELATED METHODS #
    ###########################

    @staticmethod
    def _identify_main_source(spec: rpm.spec) -> int:
        # lowest index is the main source
        return min([s[1] for s in spec.sources if s[2] == 1])

    @staticmethod
    def _get_spec_sources_list(spec_object):
        """
        Method uses RPM API to get list of Sources from the SPEC file and returns the list of sources. If the Source
        contains URL, the URL will be included in the list. This means no modifications of Sources are done at this
        point.

        :param spec_object: instance of rpm.spec object
        :type spec_object: rpm.spec
        :return: list of Sources in SPEC file in the exact order as they are listed in SPEC file.
        :rtype: list
        """
        # the sources list returned by RPM API contains list of items (path, index, source_type).
        # source type "1" is a regular source
        regular_sources = [source[:2] for source in spec_object.sources if source[2] == 1]
        regular_sources = [source[0] for source in sorted(regular_sources, key=itemgetter(1))]
        return regular_sources

    def get_sources(self) -> List[str]:
        """Gets a list of local sources."""
        return [os.path.join(self.sources_location, os.path.basename(source)) for source in self.sources]

    def get_archive(self):
        """
        Method returns the basename of first Source in SPEC file a.k.a. Source0

        :return: basename of first Source in SPEC file
        :rtype: str
        """
        return os.path.basename(self.get_sources()[0])

    def _get_raw_source_string(self, source_num: int) -> Optional[str]:
        tag = 'Source{0}'.format(source_num)
        return self.get_raw_tag_value(tag)

    def get_main_source(self) -> str:
        return self._get_raw_source_string(self.main_source_index) or ''

    ###########################
    # PATCHES RELATED METHODS #
    ###########################

    def _get_initial_patches(self) -> Dict[str, List[PatchObject]]:
        """Returns a dict of patches from a spec file"""
        patches_applied = []
        patches_not_used = []
        patches_list = [p for p in self.spc.sources if p[2] == 2]
        strip_options = self._get_patch_strip_options(patches_list)

        for patch, num, _ in patches_list:
            is_url = bool(urllib.parse.urlparse(patch).scheme)
            filename = os.path.basename(patch) if is_url else patch
            patch_path = os.path.join(self.sources_location, filename)
            if not os.path.exists(patch_path):
                if is_url:
                    logger.info('Patch%s is remote, trying to download the patch', num)
                    try:
                        DownloadHelper.download_file(patch, filename)
                    except DownloadError:
                        logger.error('Could not download remote patch %s', patch)
                        continue
                else:
                    logger.error('Patch %s does not exist', filename)
                    continue
            patch_num = num
            if patch_num in strip_options:
                patches_applied.append(PatchObject(patch_path, patch_num, strip_options[patch_num]))
            else:
                patches_not_used.append(PatchObject(patch_path, patch_num, None))
        patches_applied = sorted(patches_applied, key=lambda x: x.index)
        return {"applied": patches_applied, "not_applied": patches_not_used}

    def _get_patch_strip_options(self, patches: List[Tuple[str, int, int]]):
        """
        Gets value of strip option of each used patch

        This should work reliably in most cases except when a list of patches
        is read from a file (netcf, libvirt).
        """
        parser = SilentArgumentParser()
        parser.add_argument('-p', type=int, default=1)
        result: Dict[int, int] = {}
        for line in self.get_prep_section():
            try:
                tokens = shlex.split(line, comments=True)
            except ValueError:
                continue
            if not tokens:
                continue
            args = tokens[1:]
            try:
                ns, rest = parser.parse_known_args(args)
            except ParseError:
                continue
            rest = [os.path.basename(a) for a in rest]
            indexes = [p[1] for p in patches if os.path.basename(p[0]) in rest]
            for idx in indexes:
                if idx not in result or result[idx] < ns.p:
                    result[idx] = ns.p
        return result

    def get_patches(self):
        """
        Method returns list of all applied and not applied patches

        :return: list of PatchObject
        """
        return self.get_applied_patches() + self.get_not_used_patches()

    def get_applied_patches(self):
        """
        Method returns list of all applied patches.

        :return: list of PatchObject
        """
        return self.patches['applied']

    def get_not_used_patches(self):
        """
        Method returns list of all unpplied patches.

        :return: list of PatchObject
        """
        return self.patches['not_applied']

    def _process_patches(self, comment_out=None, remove_patches=None, disable_inapplicable_patches=None):
        """
        Comment out and delete patches from SPEC file

        :var comment_out: list with patch numbers to comment out
        :var remove_patches: list with patch numbers to delete
        :var disable_inapplicable_patches: boolean value deciding if the inapplicable patches should be commented out
        """
        if comment_out is None:
            comment_out = []
        if remove_patches is None:
            remove_patches = []

        prep = self.spec_content.section('%prep')
        if not prep:
            return

        patch_re = re.compile(r'^%patch(?P<index>\d+)(.*)')

        i = 0
        removed = 0
        while i < len(prep):
            line = prep[i]
            match = patch_re.match(line)
            if match:
                index = int(match.group('index'))
                if index in comment_out:
                    if disable_inapplicable_patches:
                        prep[i] = '#%{}'.format(line)
                        removed += 1
                    prep.insert(i, '# The following patch contains conflicts')
                    comment_out.remove(index)
                    i += 1
                elif index in remove_patches:
                    del prep[i]
                    remove_patches.remove(index)
                    i -= 1
                    removed += 1
                # When combining Patch tags and %patchlist, if a Patch is removed, the indexes
                # of %patchlist patches change and %patch macros need to be modified.
                elif self.tag('Patch{}'.format(index)).section_name.startswith('%patchlist'):
                    prep[i] = patch_re.sub(r'%patch{}\2'.format(index - removed), prep[i])
            i += 1

    @saves
    def update_paths_to_sources_and_patches(self) -> None:
        """Fixes paths of patches and sources to make them usable in SPEC file location"""
        rebased_sources_path = os.path.join(constants.RESULTS_DIR, constants.REBASED_SOURCES_DIR)
        for tag_type in ('Patch', 'Source'):
            for tag in self.tags.filter(name='{}*'.format(tag_type)):
                value = self.get_raw_tag_value(tag.name)
                if value and not urllib.parse.urlparse(value).scheme:
                    self.set_raw_tag_value(tag.name, value.replace(rebased_sources_path + os.path.sep, ''))

    @saves
    def write_updated_patches(self, patches: Dict[str, List[str]], disable_inapplicable: bool) -> None:
        """Updates SPEC file according to rebased patches.

        Args:
            patches: Dict of lists of modified, deleted or inapplicable patches.
            disable_inapplicable: Whether to comment out inapplicable patches.
        """
        def is_comment(line):
            if re.match(r'^#\s*[A-Za-z][A-Za-z0-9]+\s*:(?!//)', line):
                # ignore commented-out tag
                return False
            return line.startswith('#')

        def is_empty(line):
            return not line or line.isspace()
        if not patches:
            return None
        removed_patches = []
        inapplicable_patches = []
        modified_patches = []
        remove_lines: Dict[int, List[Tuple[int, int]]] = collections.defaultdict(list)
        for tag in self.tags.filter(name='Patch*'):
            section = self.spec_content[tag.section_index]
            if section is None:
                continue
            patch_name = os.path.basename(self.get_raw_tag_value(tag.name) or '')
            if 'deleted' in patches:
                patch_removed = [x for x in patches['deleted'] if patch_name in x]
            else:
                patch_removed = []
            if 'inapplicable' in patches:
                patch_inapplicable = [x for x in patches['inapplicable'] if patch_name in x]
            else:
                patch_inapplicable = []
            if patch_removed:
                # remove the line of the patch that was removed
                self.removed_patches.append(patch_name)
                removed_patches.append(tag.index)
                # find associated comments
                i = tag.line
                # if the tag is followed by an empty line remove empty lines
                # in front of the tag to avoid unnecessary blank lines in the spec.
                blank_follows = i + 1 < len(section) and is_empty(section[i + 1])
                while i > 0 and (is_comment(section[i - 1]) or blank_follows and is_empty(section[i - 1])):
                    i -= 1
                remove_lines[tag.section_index].append((i, tag.line + 1))
                continue
            if patch_inapplicable:
                if disable_inapplicable:
                    # comment out line if the patch was not applied
                    section[tag.line] = '#' + section[tag.line]
                inapplicable_patches.append(tag.index)
            if 'modified' in patches:
                patch = [x for x in patches['modified'] if patch_name in x]
            else:
                patch = []
            if patch:
                name = os.path.join(constants.RESULTS_DIR, constants.REBASED_SOURCES_DIR, patch_name)
                self.set_raw_tag_value(tag.name, name)
                modified_patches.append(tag.index)
        for section_index, remove in remove_lines.items():
            content = self.spec_content[section_index]
            for span in sorted(remove, key=lambda s: s[0], reverse=True):
                del content[slice(*span)]
        self._process_patches(inapplicable_patches, removed_patches, disable_inapplicable)

    ###################################
    # PACKAGE VERSION RELATED METHODS #
    ###################################

    def get_NVR(self) -> str:
        return '{0.name}-{0.version}-{0.release}'.format(self.header)

    def get_version(self) -> str:
        # deprecated, kept for backward compatibility
        return self.header.version

    def get_release(self) -> str:
        """Returns release string without %dist"""
        release = self.header.release
        dist = MacroHelper.expand('%{dist}')
        if dist and release.endswith(dist):
            release = release[:-len(dist)]
        return release

    def parse_release(self) -> Tuple[bool, int, Optional[str]]:
        """Parses release string.

        Returns:
            Tuple of is_prerelease, release_number and extra_version.

        Raises:
            RebaseHelperError in case release string is not valid.

        """
        release = self.get_release()
        m = re.match(r'^(0\.)?(\d+)(?:\.(.+))?$', release)
        if not m:
            raise RebaseHelperError('Invalid release string: {}'.format(release))
        return bool(m.group(1)), int(m.group(2)), m.group(3)

    def set_version(self, version: str) -> None:
        logger.verbose('Updating version in SPEC from %s to %s', self.header.version, version)
        self.set_tag('Version', version, preserve_macros=True)

    def set_release(self, release: str) -> None:
        logger.verbose('Changing release to %s', release)
        self.set_tag('Release', '{}%{{?dist}}'.format(release), preserve_macros=True)

    def set_release_number(self, release: str) -> None:
        # deprecated, kept for backward compatibility
        self.set_release(release)

    def set_extra_version(self, extra_version: Optional[str], version_changed: bool) -> None:
        """Updates SPEC file with the specified extra version.

        Args:
            extra_version: Extra version string or None.
            version_changed: Whether version (the value of Version tag) changed.

        """
        logger.verbose('Setting extra version in SPEC to %s', extra_version)
        relnum = self.parse_release()[1]
        relnum = 1 if version_changed else relnum + 1
        release = str(relnum)
        if extra_version:
            release += '.' + extra_version
            if re.match(r'^(a(lpha)?|b(eta)?|cr|rc)\d*$', extra_version, re.IGNORECASE):
                release = '0.' + release
        self.set_release(release)
        # TODO: in some cases it might be necessary to modify Source0

    @saves
    def set_tag(self, tag: str, value: str, preserve_macros: bool = False) -> None:
        """Sets value of a tag while trying to preserve macros if requested.

        Note that this method is not intended to be used with non-unique tags, it will only affect the first instance.

        Args:
            tag: Tag name.
            value: Tag value.
            preserve_macros: Whether to attempt to preserve macros in the current tag value.

        """
        macro_def_re = re.compile(
            r'''
            ^
            (?P<cond>%{!?\?\w+:\s*)?
            (?(cond)%global|%(global|define))
            \s+
            (?P<name>\w+)
            (?P<options>\(.+?\))?
            \s+
            (?P<value>
                (%((?P<b>{)|(?P<s>\()))?
                .+?
                (?(b)})(?(s)\))
            )
            (?(cond)})
            $
            ''',
            re.VERBOSE | re.MULTILINE | re.DOTALL)

        def _get_macro_value(macro):
            """Returns raw value of a macro"""
            for match in macro_def_re.finditer('\n'.join(self.spec_content.section('%package'))):
                if match.group('name') == macro:
                    return match.group('value')
            return None

        def _redefine_macro(macro, value):
            """Replaces value of an existing macro"""
            content = '\n'.join(self.spec_content.section('%package'))
            for match in macro_def_re.finditer(content):
                if match.group('name') != macro:
                    continue
                content = content[:match.start('value')] + value + content[match.end('value'):]
                if match.group('options'):
                    content = content[:match.start('options')] + content[match.end('options'):]
                break
            self.spec_content.replace_section('%package', content.split('\n'))
            self.save()

        def _find_macros(s):
            """Returns all redefinable macros present in a string"""
            macro_re = re.compile(r'%(?P<brace>{\??)?(?P<name>\w+)(?(brace)})')
            macros = []
            for match in macro_def_re.finditer('\n'.join(self.spec_content.section('%package'))):
                macros.append(match.group('name'))
            result = []
            for match in macro_re.finditer(s):
                if not match:
                    continue
                if match.group('name') not in macros:
                    continue
                result.append((match.group('name'), match.span()))
            return result

        def _expand_macros(s):
            """Expands all redefinable macros containing redefinable macros.

            Keeps track of all expanded macros. Returns the expanded string and a set of expanded
            macro names.
            """
            replace = []
            macros = set()
            for macro, span in _find_macros(s):
                value = _get_macro_value(macro)
                if not value:
                    continue
                macros.add(macro)
                rep, new_macros = _expand_macros(value)
                macros |= new_macros
                if _find_macros(rep):
                    replace.append((rep, span))
            for rep, span in reversed(replace):
                s = s[:span[0]] + rep + s[span[1]:]
            return s, macros

        def _tokenize(s):
            """Removes conditional macros and splits string on macro boundaries"""
            def parse(inp):
                tree = []
                text = ''
                macro = ''
                buf = ''
                escape = False
                while inp:
                    c = inp.pop(0)
                    if c == '%':
                        c = inp.pop(0)
                        if c == '%':
                            text += c
                        elif c == '{':
                            if text:
                                tree.append(('t', text))
                                text = ''
                            while inp and c not in ':}':
                                c = inp.pop(0)
                                buf += c
                            if c == ':':
                                tree.append(('c', buf[:-1], parse(inp)))
                                buf = ''
                            elif c == '}':
                                tree.append(('m', buf[:-1]))
                                buf = ''
                        elif c == '(':
                            if text:
                                tree.append(('t', text))
                                text = ''
                            tree.append(('s', None, parse(inp)))
                        else:
                            if text:
                                tree.append(('t', text))
                                text = ''
                            while inp and (c.isalnum() or c == '_'):
                                c = inp.pop(0)
                                macro += c
                            tree.append(('m', macro))
                            macro = ''
                    elif c == '$':
                        text += c
                        c = inp.pop(0)
                        if c == '{':
                            text += c
                            escape = True
                    elif c == '}':
                        if escape:
                            text += c
                            escape = False
                        else:
                            if text:
                                tree.append(('t', text))
                            inp.append(c)
                            return tree
                    elif c == ')':
                        if text:
                            tree.append(('t', text))
                        inp.append(c)
                        return tree
                    else:
                        text += c
                if text:
                    tree.append(('t', text))
                return tree

            def traverse(tree):
                result = []
                for node in tree:
                    if node[0] == 't':
                        # split text nodes on usual separators
                        result.extend([t for t in re.split(r'(\.|-|_)', node[1]) if t])
                    elif node[0] == 'm':
                        m = '%{{{}}}'.format(node[1])
                        if MacroHelper.expand(m):
                            result.append(m)
                    elif node[0] == 'c':
                        if MacroHelper.expand('%{{{}:1}}'.format(node[1])):
                            result.extend(traverse(node[2]))
                    elif node[0] == 's':
                        # ignore shell expansions, push nonsensical value
                        result.append('@')
                return result

            inp = list(s)
            tree = parse(inp)
            return traverse(tree)

        def _sync_macros(s):
            """Makes all macros present in a string up-to-date in rpm context"""
            _, macros = _expand_macros(s)
            for macro in macros:
                MacroHelper.purge_macro(macro)
                value = _get_macro_value(macro)
                if value and MacroHelper.expand(value):
                    rpm.addMacro(macro, value)

        def _process_value(curval, newval):
            """
            Replaces non-redefinable-macro parts of curval with matching parts from newval
            and redefines values of macros accordingly
            """
            value, _ = _expand_macros(curval)
            _sync_macros(curval + newval)
            tokens = _tokenize(value)
            values = [None] * len(tokens)
            sm = SequenceMatcher(a=newval)
            i = 0
            # split newval to match tokens
            for index, token in enumerate(tokens):
                if token[0] == '%':
                    # for macros, try both literal and expanded value
                    for v in [token, MacroHelper.expand(token, token)]:
                        sm.set_seq2(v)
                        m = sm.find_longest_match(i, len(newval), 0, len(v))
                        valid = m.size == len(v)  # only full match is valid
                        if valid:
                            break
                else:
                    sm.set_seq2(token)
                    m = sm.find_longest_match(i, len(newval), 0, len(token))
                    valid = m.size > 0
                if not valid:
                    continue
                if token == sm.b:
                    tokens[index] = token[m.b:m.b+m.size]
                if index > 0:
                    values[index] = newval[m.a:m.a+m.size]
                    if not values[index - 1]:
                        values[index - 1] = newval[i:m.a]
                    else:
                        values[index - 1] += newval[i:m.a]
                else:
                    values[index] = newval[i:m.a+m.size]
                i = m.a + m.size
            if newval[i:] and values:
                if not values[-1]:
                    values[-1] = newval[i:]
                else:
                    values[-1] += newval[i:]
            # try to fill empty macros
            for index, token in enumerate(tokens):
                if token[0] == '%':
                    continue
                if token == values[index]:
                    continue
                for i in range(index - 1, 0, -1):
                    if tokens[i][0] == '%' and not values[i]:
                        values[i] = values[index]
                        values[index] = None
                        break
            # try to make values of identical macros equal
            for index, token in enumerate(tokens):
                if token[0] != '%':
                    continue
                for i in range(index - 1, 0, -1):
                    if tokens[i] == token:
                        idx = values[index].find(values[i])
                        if idx >= 0:
                            prefix = values[index][:idx]
                            for j in range(index - 1, i + 1, -1):
                                # first non-macro token
                                if tokens[j][0] != '%':
                                    if prefix.endswith(values[j]):
                                        # move token from the end of prefix to the beginning
                                        prefix = values[j] + prefix[:prefix.find(values[j])]
                                    else:
                                        # no match with prefix, cannot continue
                                        break
                                else:
                                    # remove prefix from the original value and append it to the value of this macro
                                    values[index] = values[index][idx:]
                                    values[j] += prefix
                                    break
                        break
            # redefine macros and update tokens
            for index, token in enumerate(tokens):
                if token == values[index]:
                    continue
                if not values[index]:
                    values[index] = '%{nil}' if token[0] == '%' else ''
                macros = _find_macros(token)
                if macros:
                    _redefine_macro(macros[0][0], values[index])
                else:
                    tokens[index] = values[index]
            result = ''.join(tokens)
            _sync_macros(curval + result)
            # only change value if necessary
            if MacroHelper.expand(curval) == MacroHelper.expand(result):
                return curval
            return result

        if preserve_macros:
            value = _process_value(self.get_raw_tag_value(tag) or '', value)
        self.set_raw_tag_value(tag, value)

    @staticmethod
    def extract_version_from_archive_name(archive_path: str, main_source: str) -> str:
        """Extracts version string from source archive name.

        Args:
            archive_path: Path to the main sources archive.
            main_source: Value of Source0 tag.

        Returns:
            Extracted version string.

        Raises:
            RebaseHelperError in case version can't be determined.

        """
        fallback_regex = r'\w*[-_]?v?([.\d]+.*)({0})'.format(
            '|'.join([re.escape(a) for a in Archive.get_supported_archives()]))
        source = os.path.basename(main_source)
        regex = re.sub(r'%({)?version(?(1)})(.*%(\w+|{.+}))?', 'PLACEHOLDER', source, flags=re.IGNORECASE)
        regex = MacroHelper.expand(regex, regex)
        regex = re.escape(regex).replace('PLACEHOLDER', r'(.+)')
        if regex == re.escape(MacroHelper.expand(source, source)):
            # no substitution was made, use the fallback regex
            regex = fallback_regex
        logger.debug('Extracting version from archive name using %s', regex)
        archive_name = os.path.basename(archive_path)
        m = re.match(regex, archive_name)
        if m:
            logger.debug('Extracted version %s', m.group(1))
            return m.group(1)
        if regex != fallback_regex:
            m = re.match(fallback_regex, archive_name)
            if m:
                logger.debug('Extracted version %s', m.group(1))
                return m.group(1)
        raise RebaseHelperError('Unable to extract version from archive name')

    @staticmethod
    def split_version_string(version_string: str, current_version: str) -> Tuple[str, Optional[str]]:
        """Splits version string into version and extra version.

        Args:
            version_string: Complete version string.
            current_version: Current version (the value of Version tag).

        Returns:
            Tuple of version and extra_version.

        Raises:
            RebaseHelperError in case passed version string is not valid.

        """
        version_re = re.compile(r'^(\d+[.\d]*\d+|\d+)(\.|-|_|\+|~)?(\w+)?$')
        m = version_re.match(version_string)
        if not m:
            raise RebaseHelperError('Invalid version string: {}'.format(version_string))
        version, separator, extra = m.groups()
        m = version_re.match(current_version)
        if not m:
            raise RebaseHelperError('Invalid version string: {}'.format(current_version))
        if m.group(3):
            # if current version contains non-numeric characters, the new version should too
            version += (separator or '') + (extra or '')
            extra = None  # type: ignore  # the type is actually Optional[str], but is defined as str in typeshed
        logger.debug('Split version string %s into %s and %s', version_string, version, extra)
        return version, extra

    #################################
    # SPEC SECTIONS RELATED METHODS #
    #################################

    def get_prep_section(self):
        """Function returns whole prep section"""
        def unmatched_quotation(s):
            try:
                shlex.split(s, comments=True)
            except ValueError:
                return True
            return False
        if not self.prep_section:
            return []
        prep = self.prep_section.split('\n')
        # join lines split by backslash or ending with pipe
        result = [prep.pop(0)]
        while prep:
            if result[-1].rstrip().endswith('\\'):
                result[-1] = result[-1][:-1] + prep.pop(0)
            elif result[-1].rstrip().endswith('|') or unmatched_quotation(result[-1]):
                result[-1] = result[-1] + prep.pop(0)
            else:
                result.append(prep.pop(0))
        return result

    @staticmethod
    def get_subpackage_name(files_section):
        """Gets subpackage name based on the %files section."""
        parser = SilentArgumentParser()
        parser.add_argument('-n', default=None)
        parser.add_argument('-f')
        parser.add_argument('subpackage', nargs='?', default=None)
        ns, _ = parser.parse_known_args(shlex.split(files_section)[1:])
        if ns.n:
            return ns.n
        elif ns.subpackage:
            return '%{{name}}-{}'.format(ns.subpackage)
        else:
            return '%{name}'

    def get_main_files_section(self):
        """Finds the exact name of the main %files section.

        Returns:
            str: Name of the main files section.

        """
        for sec_name, _ in self.spec_content.sections:
            if sec_name.startswith('%files'):
                if self.get_subpackage_name(sec_name) == '%{name}':
                    return sec_name

    #############################################
    # SPEC CONTENT MANIPULATION RELATED METHODS #
    #############################################

    def _read_spec_content(self) -> SpecContent:
        """Reads the content of the Spec file.

        Returns:
            The created SpecContent instance.

        Raises:
            RebaseHelperError: If the Spec file cannot be read.

        """
        try:
            with open(self.path) as f:
                content = f.read()
        except IOError as e:
            raise RebaseHelperError("Unable to open and read SPEC file '{}'".format(self.path)) from e
        return SpecContent(content)

    def _write_spec_content(self):
        """Writes the current state of SpecContent into a file."""
        logger.verbose("Writing SPEC file '%s' to the disc", self.path)
        try:
            with open(self.path, "w") as f:
                f.write(str(self.spec_content))
        except IOError as e:
            raise RebaseHelperError("Unable to write updated data to SPEC file '{}'".format(self.path)) from e

    def copy(self, new_path):
        """Creates a copy of the current object and copies the SPEC file
        to a new location.

        Args:
            new_path (str): Path to copy the new SPEC file to.

        Returns:
            SpecFile: The created SpecFile instance.

        """
        shutil.copy(self.path, new_path)
        new_object = SpecFile(new_path, self.sources_location, self.predefined_macros)
        return new_object

    def reload(self):
        """Reloads the whole Spec file."""
        self.update()

    def save(self) -> None:
        """Saves changes made to SpecContent and updates the internal state."""
        self._write_spec_content()
        #  Update internal variables
        self.update()

    ####################
    # UNSORTED METHODS #
    ####################

    def is_test_suite_enabled(self):
        """
        Returns whether test suite is enabled during the build time

        :return: True if enabled or False if not
        """
        check_section = self.spec_content.section('%check')
        if not check_section:
            return False
        # Remove commented lines
        check_section = [x.strip() for x in check_section if not x.strip().startswith('#')]
        # If there is at least one line with some command in %check we assume test suite is run
        if check_section:
            return True
        else:
            return False

    @saves
    def update_changelog(self, changelog_entry: str) -> None:
        """Inserts a new entry into the changelog and saves the SpecFile.

        Args:
            changelog_entry: Message to use in the entry.

        """
        new_entry = self.get_new_log(changelog_entry)
        changelog = self.spec_content.section('%changelog')
        if changelog is None:
            changelog = []
            self.spec_content.replace_section('%changelog', changelog)
        changelog[0:0] = new_entry

    def get_new_log(self, changelog_entry):
        """Constructs a new changelog entry.

        Args:
            changelog_entry (str): Message to use in the entry.

        Returns:
            list: List of lines of the new entry.

        """
        new_record = []
        today = date.today()
        evr = '{epoch}:{ver}-{rel}'.format(epoch=self.header.epochnum,
                                           ver=self.header.version,
                                           rel=self.get_release())
        evr = evr[2:] if evr.startswith('0:') else evr
        new_record.append('* {day} {name} <{email}> - {evr}'.format(day=today.strftime('%a %b %d %Y'),
                                                                    name=GitHelper.get_user(),
                                                                    email=GitHelper.get_email(),
                                                                    evr=evr))
        self.update()
        # FIXME: ugly workaround for mysterious rpm bug causing macros to disappear
        self.update()
        new_record.append(MacroHelper.expand(changelog_entry, changelog_entry))
        new_record.append('')
        return new_record

    def _get_setup_parser(self):
        """
        Construct ArgumentParser for parsing %(auto)setup macro arguments

        :return: constructed ArgumentParser
        """
        parser = SilentArgumentParser()
        parser.add_argument('-n', default=MacroHelper.expand('%{name}-%{version}', '%{name}-%{version}'))
        parser.add_argument('-a', type=int, default=-1)
        parser.add_argument('-b', type=int, default=-1)
        parser.add_argument('-T', action='store_true')
        parser.add_argument('-q', action='store_true')
        parser.add_argument('-c', action='store_true')
        parser.add_argument('-D', action='store_true')
        parser.add_argument('-v', action='store_true')
        parser.add_argument('-N', action='store_true')
        parser.add_argument('-p', type=int, default=-1)
        parser.add_argument('-S', default='')
        return parser

    def get_setup_dirname(self):
        """
        Get dirname from %setup or %autosetup macro arguments

        :return: dirname
        """
        parser = self._get_setup_parser()

        prep = self.spec_content.section('%prep')
        if not prep:
            return None

        for line in prep:
            if line.startswith('%setup') or line.startswith('%autosetup'):
                args = shlex.split(line)
                args = [MacroHelper.expand(a, '') for a in args[1:]]

                # parse macro arguments
                try:
                    ns, _ = parser.parse_known_args(args)
                except ParseError:
                    continue

                # check if this macro instance is extracting Source0
                if not ns.T or ns.a == 0 or ns.b == 0:
                    return ns.n

        return None

    @saves
    def update_setup_dirname(self, dirname):
        """
        Update %setup or %autosetup dirname argument if needed

        :param dirname: new dirname to be used
        """
        parser = self._get_setup_parser()

        prep = self.spec_content.section('%prep')
        if not prep:
            return

        for index, line in enumerate(prep):
            if line.startswith('%setup') or line.startswith('%autosetup'):
                args = shlex.split(line)
                macro = args[0]
                args = [MacroHelper.expand(a, '') for a in args[1:]]

                # parse macro arguments
                try:
                    ns, unknown = parser.parse_known_args(args)
                except ParseError:
                    continue

                # check if this macro instance is extracting Source0
                if ns.T and ns.a != 0 and ns.b != 0:
                    continue

                # check if modification is really necessary
                if dirname != ns.n:
                    new_dirname = dirname

                    # get %{name} and %{version} macros
                    macros = [m for m in MacroHelper.filter(self.macros, level=-3) if m['name'] in ('name', 'version')]
                    # add all macros from spec file scope
                    macros.extend(MacroHelper.filter(self.macros, level=0))
                    # omit short macros
                    macros = [m for m in macros if len(m['value']) > 1]
                    # ensure maximal greediness
                    macros.sort(key=lambda k: len(k['value']), reverse=True)

                    # substitute tokens with macros
                    for m in macros:
                        if m['value'] and m['value'] in dirname:
                            new_dirname = new_dirname.replace(m['value'], '%{{{}}}'.format(m['name']))

                    args = [macro]
                    args.extend(['-n', new_dirname])
                    if ns.a != -1:
                        args.extend(['-a', str(ns.a)])
                    if ns.b != -1:
                        args.extend(['-b', str(ns.b)])
                    if ns.T:
                        args.append('-T')
                    if ns.q:
                        args.append('-q')
                    if ns.c:
                        args.append('-c')
                    if ns.D:
                        args.append('-D')
                    if ns.v:
                        args.append('-v')
                    if ns.N:
                        args.append('-N')
                    if ns.p != -1:
                        args.extend(['-p', str(ns.p)])
                    if ns.S != '':
                        args.extend(['-S', ns.S])
                    args.extend(unknown)

                    prep[index] = ' '.join(args)

    def find_archive_target_in_prep(self, archive):
        """
        Tries to find a command that is used to extract the specified archive
        and attempts to determine target path from it.
        'tar' and 'unzip' commands are supported so far.

        :param archive: Path to archive
        :return: Target path relative to builddir or None if not determined
        """
        cd_parser = SilentArgumentParser()
        cd_parser.add_argument('dir', default=os.environ.get('HOME', ''))
        tar_parser = argparse.ArgumentParser()
        tar_parser.add_argument('-C', default='.', dest='target')
        unzip_parser = argparse.ArgumentParser()
        unzip_parser.add_argument('-d', default='.', dest='target')
        archive = os.path.basename(archive)
        builddir = MacroHelper.expand('%{_builddir}', '')
        basedir = builddir
        for line in self.get_prep_section():
            tokens = shlex.split(line, comments=True)
            if not tokens:
                continue
            # split tokens by pipe
            for tokens in [list(group) for k, group in itertools.groupby(tokens, lambda t: t == '|') if not k]:
                cmd, args = os.path.basename(tokens[0]), tokens[1:]
                if cmd == 'cd':
                    # keep track of current directory
                    try:
                        ns, _ = cd_parser.parse_known_args(args)
                    except ParseError:
                        pass
                    else:
                        basedir = ns.dir if os.path.isabs(ns.dir) else os.path.join(basedir, ns.dir)
                if archive in line:
                    if cmd == 'tar':
                        parser = tar_parser
                    elif cmd == 'unzip':
                        parser = unzip_parser
                    else:
                        continue
                    try:
                        ns, _ = parser.parse_known_args(args)
                    except ParseError:
                        continue
                    basedir = os.path.relpath(basedir, builddir)
                    return os.path.normpath(os.path.join(basedir, ns.target))
        return None



import re
import textwrap

from docutils import nodes
from sphinx.writers.manpage import ManualPageTranslator
from sphinx.builders.manpage import ManualPageBuilder


class CustomManPageTranslator(ManualPageTranslator):
    """
    Manual page translator which properly formats option text roles
    and option directives
    """
    def __init__(self, builder, *args, **kwds):
        ManualPageTranslator.__init__(self, builder, *args, **kwds)
        self.option_context = []

    @staticmethod
    def _format_option(text):
        """
        Format single option, make option strings bold and option arguments italic
        """
        def get_font(token):
            # option strings start with "-"
            return 'B' if token['text'].startswith('-') else 'I'
        # find all tokens to be decorated, excluding "..."
        matches = list(re.finditer(r'[^\s\[\]{}=,|]+', text))
        tokens = [dict(text=m.group(0), span=m.span()) for m in matches if m.group(0) != '...']
        for t in reversed(tokens):
            text = text[: t['span'][1]] + '\\fP' + text[t['span'][1]:]
            text = text[: t['span'][0]] + '\\f{}'.format(get_font(t)) + text[t['span'][0]:]
        return text

    def visit_desc_signature(self, node):
        self.visit_definition_list_item(node)
        self.body.append('\n')

    def depart_desc_signature(self, node):
        # format option text formed by preceding desc_names and desc_addnames
        if self.option_context:
            self.body.append(self._format_option(''.join(self.option_context)))
            del self.option_context[:]
        self.body.append('\n')

    def visit_desc_name(self, node):
        # save text for later
        self.option_context.append(node.astext())
        raise nodes.SkipNode

    def visit_desc_addname(self, node):
        text = node.astext()
        # desc_addname node containing only "," is taken as option separator
        if text.strip() == ',':
            # format option text formed by preceding desc_names and desc_addnames
            if self.option_context:
                self.body.append(self._format_option(''.join(self.option_context)))
                del self.option_context[:]
        else:
            # save text for later
            self.option_context.append(text)
            raise nodes.SkipNode

    def visit_reference(self, node):
        if node.children and isinstance(node[0], nodes.literal):
            if 'std-option' in node[0].get('classes', ''):
                # prevent ignoring nested option node
                raise nodes.SkipDeparture
        ManualPageTranslator.visit_reference(self, node)

    def visit_literal(self, node):
        if 'std-option' in node.get('classes', ''):
            try:
                max_width = self.settings.env.autoargs_options['synopsis_max_width']
            except (AttributeError, KeyError):
                text = node.astext()
            else:
                # wrap lines exceeding max width
                text = '\n'.join(textwrap.wrap(node.astext(), width=max_width))
            # format option text
            self.body.append(self._format_option(text))
            raise nodes.SkipNode
        else:
            self.visit_emphasis(node)

    def depart_literal(self, node):
        self.depart_emphasis(node)

    def visit_manpage(self, node):
        text = node.astext()
        # parentheses and section numbers inside them shouldn't be bold
        text = re.sub(r'(\(\d*\))', r'\\fR\1\\fP', text)
        self.body.append('\\fB{}\\fP'.format(text))
        raise nodes.SkipNode


class CustomManPageBuilder(ManualPageBuilder):
    """Manual page builder which uses CustomManualPageTranslator"""
    name = 'custom-man'
    format = 'man'
    # use custom translator (Sphinx >= 1.6)
    default_translator_class = CustomManPageTranslator

    def init(self):
        ManualPageBuilder.init(self)
        # use custom translator (Sphinx < 1.6)
        try:
            self.translator_class = CustomManPageTranslator
        except AttributeError:
            pass


def setup(app):
    app.add_builder(CustomManPageBuilder)
    return dict(version='0.1')



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import logging
import re
from typing import cast

from rebasehelper.logger import CustomLogger
from rebasehelper.plugins.spec_hooks import BaseSpecHook
from rebasehelper.helpers.download_helper import DownloadHelper


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


class CommitHashUpdater(BaseSpecHook):
    """Tries to update commit hash present in Source0 tag according to the new version"""

    @classmethod
    def _get_commit_hash_from_github(cls, spec_file):
        """
        Tries to find a commit using Github API

        :param spec_file: SPEC file to base the search on
        :return: SHA of a commit, or None
        """
        m = re.match(r'^https?://github\.com/(?P<owner>[\w-]+)/(?P<project>[\w-]+)/.*$', spec_file.sources[0])
        if not m:
            return None
        baseurl = 'https://api.github.com/repos/{owner}/{project}'.format(**m.groupdict())
        # try to get tag name from a release matching version
        r = DownloadHelper.request('{}/releases'.format(baseurl))
        if r is None:
            return None

        if not r.ok:
            if r.status_code == 403 and r.headers.get('X-RateLimit-Remaining') == '0':
                logger.warning("Rate limit exceeded on Github API! Try again later.")
            return None
        data = r.json()
        version = spec_file.header.version
        tag_name = None
        for release in data:
            if version in release.get('name', ''):
                tag_name = release.get('tag_name')
                break

        r = DownloadHelper.request('{}/tags'.format(baseurl))
        if r is None:
            return None
        if not r.ok:
            if r.status_code == 403 and r.headers.get('X-RateLimit-Remaining') == '0':
                logger.warning("Rate limit exceeded on Github API! Try again later.")
            return None
        data = r.json()
        for tag in data:
            name = tag.get('name')
            if tag_name:
                if name != tag_name:
                    continue
            else:
                # no specific tag name, try common tag names
                if name not in [version, 'v{}'.format(version)]:
                    continue
            commit = tag.get('commit')
            if commit:
                return commit.get('sha')
        return None

    @classmethod
    def _get_commit_hash(cls, spec_file):
        if 'github.com' in spec_file.sources[0]:
            return cls._get_commit_hash_from_github(spec_file)
        return None

    @classmethod
    def run(cls, spec_file, rebase_spec_file, **kwargs):
        if rebase_spec_file.sources[0] != spec_file.sources[0]:
            # nothing to do
            return
        # try to determine commit hash matching the new version
        new_commit = cls._get_commit_hash(rebase_spec_file)
        if not new_commit:
            return
        source = rebase_spec_file.sources[0]
        # try to determine commit hash matching the old version
        old_commit = cls._get_commit_hash(spec_file)
        if old_commit:
            # replace old commit hash with the new one
            source = source.replace(old_commit, new_commit)
        else:
            # try to find anything resembling SHA1 hash
            hashes = re.findall(r'[0-9a-f]{40}', source)
            if len(set(hashes)) != 1:
                # multiple different hashes (or none), cannot continue
                return
            source = source.replace(hashes[0], new_commit)
        tag = 'Source{0}'.format(rebase_spec_file.main_source_index)
        rebase_spec_file.set_tag(tag, source, preserve_macros=True)



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import shlex

from rebasehelper.plugins.plugin import Plugin
from rebasehelper.plugins.plugin_collection import PluginCollection


class SRPMBuildToolBase(Plugin):
    """SRPM build tool base class.

    Attributes:
        DEFAULT(bool): If True, the build tool is default tool.

    """

    DEFAULT: bool = False

    @staticmethod
    def get_srpm_builder_options(**kwargs):
        srpm_builder_options = kwargs.get('srpm_builder_options')
        if srpm_builder_options:
            return shlex.split(srpm_builder_options)
        return None

    @classmethod
    def build(cls, spec, results_dir, **kwargs):
        """
        Build SRPM with chosen SRPM Build Tool

        :param spec: SpecFile object
        :param results_dir: absolute path to DIR where results should be stored
        :return: absolute path to SRPM, list with absolute paths to logs
        """
        raise NotImplementedError()


class SRPMBuildToolCollection(PluginCollection):
    """Collection of SRPM build tools."""



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import os

import pytest  # type: ignore

from typing import List

from rebasehelper.helpers.lookaside_cache_helper import LookasideCacheHelper


class TestLookasideCacheHelper:

    TEST_FILES: List[str] = [
        'documentation.tar.xz',
        'archives/archive.tar.bz2',
    ]

    @pytest.mark.parametrize('package, filename, hashtype, hsh', [
        ('vim-go', 'v1.6.tar.gz', 'md5', '847d3e3577982a9515ad0aec6d5111b2'),
        ('rebase-helper', '0.8.0.tar.gz', 'md5', '91de540caef64cb8aa7fd250f2627a93'),
        (
                'man-pages',
                'man-pages-posix-2013-a.tar.xz',
                'sha512',
                'e6ec8eb57269fadf368aeaac31b5a98b9c71723d4d5cc189f9c4642d6e865c88'
                'e44f77481dccbdb72e31526488eb531f624d455016361687a834ccfcac19fa14',
        ),
    ], ids=[
        'vim-go',
        'rebase-helper',
        'man-pages',
    ])
    @pytest.mark.integration
    def test_download(self, package, filename, hashtype, hsh):
        # pylint: disable=protected-access
        target = os.path.basename(filename)
        LookasideCacheHelper._download_source('fedpkg',
                                              'https://integration:4430/pkgs',
                                              package,
                                              filename,
                                              hashtype,
                                              hsh,
                                              target)
        assert os.path.isfile(target)
        assert LookasideCacheHelper._hash(target, hashtype) == hsh

    @pytest.mark.parametrize('filename, hashtype, hsh', [
        ('documentation.tar.xz', 'md5', '03a77b3e59deec24c1d70a495e41602b'),
        (
                'archive.tar.bz2',
                'sha512',
                '6bab9c2cc6b73fbba27be45c6b5dc57a0d763e12e6a71bcc9fbdde61611ccaed'
                'f4474a09dc6e4f65e267f12ecd6d314ab87e6a43f1e62ea7d124720903e40eb4',
        ),
    ], ids=[
        'documentation.tar.xz',
        'archive.tar.bz2',
    ])
    @pytest.mark.integration
    def test_upload(self, filename, hashtype, hsh):
        # pylint: disable=protected-access
        LookasideCacheHelper._upload_source('https://integration:4430/pkgs',
                                            'test',
                                            '',
                                            filename,
                                            hashtype,
                                            hsh,
                                            None)



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import configparser
import os

from rebasehelper.options import OPTIONS, traverse_options
from rebasehelper.constants import CONFIG_PATH, CONFIG_FILENAME
from rebasehelper.plugins.plugin_manager import plugin_manager


class Config:

    def __init__(self, config_file=None):
        self.path_to_config = self.get_config_path(config_file)
        self.config = self.get_config()

    def __getattr__(self, name):
        return self.config.get(name)

    @staticmethod
    def get_config_path(config_file):
        # ensure XDG_CONFIG_HOME is set
        if 'XDG_CONFIG_HOME' not in os.environ:
            os.environ['XDG_CONFIG_HOME'] = os.path.expandvars(os.path.join('$HOME', '.config'))
        path = os.path.expandvars(config_file or os.path.join(CONFIG_PATH, CONFIG_FILENAME))
        return os.path.abspath(path)

    def get_config(self):
        conf = {}
        if os.path.isfile(self.path_to_config):
            config = configparser.ConfigParser()
            config.read(self.path_to_config)
            for section in config.sections():
                conf.update({k.replace('-', '_'): v for k, v in config.items(section)})

        return conf

    def merge(self, cli):
        self.config.update(vars(cli.args))

        for option in traverse_options(OPTIONS + plugin_manager.get_options()):
            if 'dest' in option:
                dest = option['dest']
            else:
                args = [n.lstrip('-').replace('-', '_') for n in option['name'] if n.startswith('--')]
                if args:
                    dest = args[0]

            if dest and dest not in self.config:
                self.config[dest] = option.get('default')



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import re

from rebasehelper.plugins.spec_hooks import BaseSpecHook


class EscapeMacros(BaseSpecHook):
    """Spec hook escaping RPM macros in comments."""

    @classmethod
    def run(cls, spec_file, rebase_spec_file, **kwargs):
        for sec_name, section in rebase_spec_file.spec_content.sections:
            for index, line in enumerate(section):
                start, end = spec_file.spec_content.get_comment_span(line, sec_name)
                new_comment = re.sub(r'(?<!%)(%(?P<brace>{\??)?\w+(?(brace)}))', r'%\1', line[start:end])
                section[index] = line[:start] + new_comment



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import json
import os

import git  # type: ignore
import pytest  # type: ignore
import unidiff  # type: ignore

from typing import List

from rebasehelper.cli import CLI
from rebasehelper.config import Config
from rebasehelper.application import Application
from rebasehelper.constants import RESULTS_DIR, CHANGES_PATCH
from rebasehelper.helpers.git_helper import GitHelper


@pytest.fixture
def initialized_git_repo(workdir):
    repo = git.Repo.init(workdir)
    # Configure user otherwise app.apply_changes() will fail
    repo.git.config('user.name', GitHelper.get_user(), local=True)
    repo.git.config('user.email', GitHelper.get_email(), local=True)
    repo.git.add(all=True)
    repo.index.commit('Initial commit', skip_hooks=True)
    return repo


class TestRebase:

    TEST_FILES: List[str] = [
        'rebase/test.spec',
        'rebase/applicable.patch',
        'rebase/backported.patch',
        'rebase/conflicting.patch',
        'rebase/renamed-0.1.patch',
    ]

    @pytest.mark.parametrize('buildtool', [
        pytest.param('rpmbuild', marks=pytest.mark.skipif(
            os.geteuid() != 0,
            reason='requires superuser privileges')),
        pytest.param('mock', marks=pytest.mark.long_running),
    ])
    @pytest.mark.parametrize('favor_on_conflict', ['upstream', 'downstream', 'off'])
    @pytest.mark.integration
    @pytest.mark.usefixtures('initialized_git_repo')
    def test_rebase(self, buildtool, favor_on_conflict):
        new_version = '0.2'

        cli = CLI([
            '--non-interactive',
            '--disable-inapplicable-patches',
            '--buildtool', buildtool,
            '--favor-on-conflict', favor_on_conflict,
            '--outputtool', 'json',
            '--pkgcomparetool', 'rpmdiff,pkgdiff,abipkgdiff,licensecheck,sonamecheck',
            '--color=always',
            '--apply-changes',
            new_version,
        ])
        config = Config()
        config.merge(cli)
        execution_dir, results_dir = Application.setup(config)
        app = Application(config, os.getcwd(), execution_dir, results_dir)
        app.run()
        changes = os.path.join(RESULTS_DIR, CHANGES_PATCH)
        patch = unidiff.PatchSet.from_filename(changes, encoding='UTF-8')

        if favor_on_conflict == 'upstream':
            backported_patch, conflicting_patch, renamed_patch, spec_file = patch
            assert conflicting_patch.is_removed_file  # conflicting.patch
        elif favor_on_conflict == 'downstream':
            backported_patch, conflicting_patch, renamed_patch, spec_file = patch
            assert conflicting_patch.is_modified_file  # conflicting.patch
        else:
            backported_patch, renamed_patch, spec_file = patch
            # Non interactive mode - inapplicable patches are only commented out.
            assert [h for h in spec_file if '+#Patch1:         conflicting.patch\n' in h.target]
            assert [h for h in spec_file if '+#%%patch1 -p1\n' in h.target]
        assert renamed_patch.is_rename  # renamed patch 0.1.patch to 0.2.patch
        assert os.path.basename(renamed_patch.source_file) == 'renamed-0.1.patch'
        assert os.path.basename(renamed_patch.target_file) == 'renamed-0.2.patch'
        # Check that the renamed patch path is unchanged
        assert not [h for h in spec_file if '-Patch3:         renamed-%{version}.patch\n' in h.source]
        assert backported_patch.is_removed_file   # backported.patch
        assert spec_file.is_modified_file  # test.spec
        if favor_on_conflict != 'downstream':
            assert [h for h in spec_file if '-Patch1:         conflicting.patch\n' in h.source]
            assert [h for h in spec_file if '-%patch1 -p1\n' in h.source]
        assert [h for h in spec_file if '-Patch2:         backported.patch\n' in h.source]
        assert [h for h in spec_file if '-%patch2 -p1\n' in h.source]
        assert [h for h in spec_file if '+- New upstream release {}\n'.format(new_version) in h.target]
        with open(os.path.join(RESULTS_DIR, 'report.json')) as f:
            report = json.load(f)
            assert 'success' in report['result']
            # patches
            assert 'applicable.patch' in report['patches']['untouched']
            if favor_on_conflict == 'upstream':
                # In case of conflict, upstream code is favored, therefore conflicting patch is unused.
                assert 'conflicting.patch' in report['patches']['deleted']
            elif favor_on_conflict == 'downstream':
                assert 'conflicting.patch' in report['patches']['modified']
            else:
                # Non interactive mode - skipping conflicting patches
                assert 'conflicting.patch' in report['patches']['inapplicable']
            assert 'backported.patch' in report['patches']['deleted']
            # licensecheck
            assert report['checkers']['licensecheck']['license_changes']
            assert len(report['checkers']['licensecheck']['disappeared_licenses']) == 1
            assert len(report['checkers']['licensecheck']['new_licenses']) == 1
            # rpmdiff
            assert report['checkers']['rpmdiff']['files_changes']['added'] == 1
            assert report['checkers']['rpmdiff']['files_changes']['changed'] == 3
            assert report['checkers']['rpmdiff']['files_changes']['removed'] == 1
            # abipkgdiff
            assert report['checkers']['abipkgdiff']['abi_changes']
            lib = report['checkers']['abipkgdiff']['packages']['test']['libtest1.so']
            if 'Function symbols changes summary' in lib:
                assert lib['Function symbols changes summary']['Added']['count'] == 1
            elif 'Functions changes summary' in lib:
                assert lib['Functions changes summary']['Added']['count'] == 1
            if favor_on_conflict != 'downstream':
                if 'Variable symbols changes summary' in lib:
                    assert lib['Variable symbols changes summary']['Removed']['count'] == 1
                elif 'Variables changes summary' in lib:
                    assert lib['Variables changes summary']['Removed']['count'] == 1
            # sonamecheck
            change = report['checkers']['sonamecheck']['soname_changes']['test']['changed'][0]
            assert change['from'] == 'libtest2.so.0.1'
            assert change['to'] == 'libtest2.so.0.2'

        repo = git.Repo(execution_dir)
        assert '- New upstream release {}'.format(new_version) in repo.commit().summary

    @pytest.mark.parametrize('buildtool', [
        pytest.param('rpmbuild', marks=pytest.mark.skipif(
            os.geteuid() != 0,
            reason='requires superuser privileges')),
        pytest.param('mock', marks=pytest.mark.long_running),
    ])
    @pytest.mark.integration
    @pytest.mark.usefixtures('initialized_git_repo')
    def test_files_build_log_hook(self, buildtool):
        new_version = '0.3'

        cli = CLI([
            '--non-interactive',
            '--disable-inapplicable-patches',
            '--force-build-log-hooks',
            '--buildtool', buildtool,
            '--outputtool', 'json',
            '--pkgcomparetool', '',
            '--color=always',
            new_version,
        ])
        config = Config()
        config.merge(cli)
        execution_dir, results_dir = Application.setup(config)
        app = Application(config, os.getcwd(), execution_dir, results_dir)
        app.run()
        changes = os.path.join(RESULTS_DIR, CHANGES_PATCH)
        patch = unidiff.PatchSet.from_filename(changes, encoding='UTF-8')

        _, _, spec_file = patch

        assert spec_file.is_modified_file
        # removed files
        assert [h for h in spec_file if '-%doc README.md CHANGELOG.md\n' in h.source]
        assert [h for h in spec_file if '+%doc README.md\n' in h.target]
        assert [h for h in spec_file if '-%doc %{_docdir}/%{name}/notes.txt\n' in h.source]
        assert [h for h in spec_file if '-%{_datadir}/%{name}/1.dat\n' in h.source]
        assert [h for h in spec_file if '-%{_datadir}/%{name}/extra/C.dat\n' in h.source]
        assert [h for h in spec_file if '-%doc data/extra/README.extra\n' in h.source]
        # added files
        assert [h for h in spec_file if '+%{_datadir}/%{name}/2.dat\n' in h.target]
        assert [h for h in spec_file if '+%{_datadir}/%{name}/extra/D.dat\n' in h.target]

        with open(os.path.join(RESULTS_DIR, 'report.json')) as f:
            report = json.load(f)
            assert 'success' in report['result']
            # files build log hook
            added = report['build_log_hooks']['files']['added']
            assert '%{_datadir}/%{name}/2.dat' in added['%files']
            assert '%{_datadir}/%{name}/extra/D.dat' in added['%files extra']
            removed = report['build_log_hooks']['files']['removed']
            assert 'CHANGELOG.md' in removed['%files']
            assert '%{_docdir}/%{name}/notes.txt' in removed['%files']
            assert '%{_datadir}/%{name}/1.dat' in removed['%files']
            assert '%{_datadir}/%{name}/extra/C.dat' in removed['%files extra']
            assert 'data/extra/README.extra' in removed['%files extra']



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import configparser
import os

import pytest  # type: ignore

from rebasehelper.cli import CLI
from rebasehelper.config import Config


class TestConfig:
    CONFIG_FILE: str = 'test_config.cfg'

    @pytest.fixture
    def config_file(self, config_args):
        config = configparser.ConfigParser()
        config.add_section('Section1')
        for key, value in config_args.items():
            config.set('Section1', key, value)
        with open(self.CONFIG_FILE, 'w') as configfile:
            config.write(configfile)

        return os.path.abspath(self.CONFIG_FILE)

    @pytest.mark.parametrize('config_args', [
        {
            'changelog-entry': 'Updated to',
            'versioneer': 'pypi',
        },
        {},
    ], ids=[
        'configured',
        'empty',
    ])
    def test_get_config(self, config_args, config_file):
        config = Config(config_file)
        expected_result = {k.replace('-', '_'): v for k, v in config_args.items()}
        assert expected_result == config.config

    @pytest.mark.parametrize('cli_args, config_args, merged', [
        (
            [
                '--changelog-entry', 'Version set to',
                '--buildtool', 'rpmbuild',
            ],
            {
                'changelog-entry': 'Updated to ',
                'versioneer': 'pypi',
            },
            {
                'changelog-entry': 'Version set to',
                'versioneer': 'pypi',
                'buildtool': 'rpmbuild',
                'color': 'auto',
            },
        ),
    ], ids=[
            'CLI with config',
    ])
    def test_merge(self, cli_args, merged, config_file):
        expected_result = {k.replace('-', '_'): v for k, v in merged.items()}
        cli = CLI(cli_args)
        config = Config(config_file)
        config.merge(cli)
        # True if expected_result is a subset of conf.config
        assert expected_result.items() <= config.config.items()



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import os
import sys

import pytest  # type: ignore

from rebasehelper.helpers.console_helper import ConsoleHelper


class TestConsoleHelper:

    def test_capture_output(self):
        def write():
            with os.fdopen(sys.__stdout__.fileno(), 'w') as f:  # pylint: disable=no-member
                f.write('test stdout')
            with os.fdopen(sys.__stderr__.fileno(), 'w') as f:  # pylint: disable=no-member
                f.write('test stderr')

        with ConsoleHelper.Capturer(stdout=True, stderr=True) as capturer:
            write()

        assert capturer.stdout == 'test stdout'
        assert capturer.stderr == 'test stderr'

    @pytest.mark.parametrize('specification, expected_rgb, expected_bit_width', [
        ('rgb:0000/0000/0000', (0x0, 0x0, 0x0), 16),
        ('rgb:ffff/ffff/ffff', (0xffff, 0xffff, 0xffff), 16),
        ('rgb:f/f/f', (0xf, 0xf, 0xf), 4),
        ('rgb:', None, None)
    ], ids=[
        '16-bit-black',
        '16-bit-white',
        '4-bit-white',
        'invalid-format',
    ])
    def test_parse_rgb_device_specification(self, specification, expected_rgb, expected_bit_width):
        rgb, bit_width = ConsoleHelper.parse_rgb_device_specification(specification)
        assert rgb == expected_rgb
        assert bit_width == expected_bit_width

    @pytest.mark.parametrize('rgb_tuple, bit_width, expected_result', [
        ((0xf, 0xf, 0xf), 4, True),
        ((0x0, 0x0, 0x0), 4, False),
        ((0x2929, 0x2929, 0x2929), 16, False),
    ], ids=[
        'white',
        'black',
        'grey',
    ])
    def test_color_is_light(self, rgb_tuple, bit_width, expected_result):
        assert ConsoleHelper.color_is_light(rgb_tuple, bit_width) == expected_result



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

from rebasehelper.cli import CLI


class TestCLI:
    def test_cli_unit(self):
        """Function tests cli class with all arguments"""
        conf = {
            'verbose': True,
            'srpm_buildtool': 'mock',
            'buildtool': 'rpmbuild',
            'outputtool': 'json',
            'keep_workspace': True,
            'not_download_sources': True,
            'non_interactive': True,
            'builds_nowait': True,
            'build_tasks': ['123456', '654321'],
            'results_dir': '/tmp/rebase-helper',
            'srpm_builder_options': '\"-r fedora-26-x86_64\"',
            'builder_options': '\"-v\"',
            'changelog_entry': 'Update to %{version}',
            'sources': 'test-1.0.3.tar.gz',
            'version': False,
            'versioneer': None,
            'disable_inapplicable_patches': False,
            'get_old_build_from_koji': False,
            'color': 'auto',
        }
        arguments = [
            '--verbose',
            '--srpm-buildtool', 'mock',
            '--buildtool', 'rpmbuild',
            '--outputtool', 'json',
            '--keep-workspace',
            '--not-download-sources',
            '--non-interactive',
            '--builds-nowait',
            '--build-tasks', '123456,654321',
            '--results-dir', '/tmp/rebase-helper',
            '--srpm-builder-options=\"-r fedora-26-x86_64\"',
            '--builder-options=\"-v\"',
            '--changelog-entry', 'Update to %{version}',
            'test-1.0.3.tar.gz',
        ]
        cli = CLI(arguments)
        for key, value in cli.args.__dict__.items():
            assert value == conf[key]



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import logging
from typing import cast

from rebasehelper.logger import CustomLogger
from rebasehelper.plugins.plugin import Plugin
from rebasehelper.plugins.plugin_collection import PluginCollection
from rebasehelper.types import PackageCategories


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


class BaseVersioneer(Plugin):
    """Base class for a versioneer"""

    CATEGORIES: PackageCategories = []

    @classmethod
    def run(cls, package_name):
        """
        Runs a versioneer.

        :param package_name: Name of a package
        :return: Latest upstream version of a package
        """
        raise NotImplementedError()


class VersioneerCollection(PluginCollection):

    def run(self, versioneer, package_name, category, versioneer_blacklist=None):
        """Runs the specified versioneer or all versioneers subsequently
        until one of them succeeds.

        Args:
            versioneer (str): Name of a versioneer.
            package_name (str): Name of a package.
            category (str): Package category.
            versioneer_blacklist (list): List of versioneers that will be skipped.

        Returns:
            str: Latest upstream version of a package.

        """
        if versioneer_blacklist is None:
            versioneer_blacklist = []

        if versioneer:
            logger.info("Running '%s' versioneer", versioneer)
            return self.plugins[versioneer].run(package_name)
        # run all versioneers, except those disabled in config, categorized first
        allowed_versioneers = [v for k, v in self.plugins.items() if v and k not in versioneer_blacklist]
        for versioneer in sorted(allowed_versioneers, key=lambda v: not v.CATEGORIES):
            categories = versioneer.CATEGORIES
            if not categories or category in categories:
                logger.info("Running '%s' versioneer", versioneer.name)
                result = versioneer.run(package_name)
                if result:
                    return result
        return None



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import collections
import re
import urllib.parse
from typing import Any, Dict, List, Pattern, Tuple, Set

from rebasehelper.exceptions import RebaseHelperError
from rebasehelper.types import Options
from rebasehelper.plugins.spec_hooks import BaseSpecHook
from rebasehelper.specfile import SpecFile
from rebasehelper.helpers.macro_helper import MacroHelper


class ReplaceOldVersion(BaseSpecHook):
    """SpecHook for replacing occurrences of old version string."""

    OPTIONS: Options = [
        {
            "name": ["--replace-old-version-with-macro"],
            "default": False,
            "switch": True,
            "help": "replace old version string with %%{version} instead of new version string",
        },
    ]

    # taken from lib/rpmte.c in RPM source
    IGNORED_TAGS = [
        # dependency tags could be specified with a version equal to the old version
        # which could break the functionality, ignore them
        'Conflicts',
        'Enhances',
        'Obsoletes',
        'Provides',
        'Recommends',
        'Requires',
        'Suggests',
        # ignore Version tag to avoid mistakes when version format changes
        # e.g. in a rebase from 2.5 to 2.5.1, version would be changed to 2.5.1.1
        'Version',
    ]

    @classmethod
    def _create_possible_replacements(cls, spec_file: SpecFile, rebase_spec_file: SpecFile,
                                      use_macro: bool) -> List[Tuple[Pattern[str], str]]:
        """Creates possible subversion replacements.

        Args:
            spec_file: Old SpecFile.
            rebase_spec_file: New SpecFile.
            use_macro: Whether %{version} macro should be used as a replacement.

        Returns:
            List of tuples containing regex pattern and replacement
            that can be passed to re.sub. The first tuple always
            represents the whole version.

        Example:
            Subversions 1.2, 1.2.3, 1.2.3.4 would be created from version 1.2.3.4.

        """
        old = spec_file.get_version()
        new = rebase_spec_file.get_version()
        version_re = r'([\ /\-\s]){}([/.\-\s]|$)'
        # Allow any character after whole version to replace strings such as
        # 1.0.1bc1
        full_version_re = r'([\ /\-\s]){}(.*)'
        replacement = r'\g<1>{}\g<2>'
        split_version = old.split('.')

        res = [
            (re.compile(full_version_re.format(re.escape(old))), replacement.format('%{version}' if use_macro else new))
        ]
        # iterate backwards to go from longer to shorter subversions
        for i in reversed(range(2, len(split_version))):
            pattern = re.compile(version_re.format(re.escape('.'.join(split_version[:i]))))
            new_subversion = replacement.format('.'.join(new.split('.')[:i]))
            res.append((pattern, new_subversion))
        # add hardcoded extra version replacement
        try:
            old_extra = spec_file.parse_release()[2]
            if old_extra:
                new_extra = rebase_spec_file.parse_release()[2] or ''
                # allow extraversion to immediately follow version
                extraversion = re.compile(r'([\ /\-\s\d\}])' + re.escape(old_extra) + r'([/.\-\s]|$)')
                res.append((extraversion, replacement.format(new_extra)))
        except RebaseHelperError:
            # silently skip unparsable release
            pass
        return res

    @classmethod
    def run(cls, spec_file: SpecFile, rebase_spec_file: SpecFile, **kwargs: Any):
        replace_with_macro = bool(kwargs.get('replace_old_version_with_macro'))

        subversion_patterns = cls._create_possible_replacements(spec_file, rebase_spec_file, replace_with_macro)
        examined_lines: Dict[int, Set[int]] = collections.defaultdict(set)
        for tag in rebase_spec_file.tags.filter():
            examined_lines[tag.section_index].add(tag.line)
            value = rebase_spec_file.get_raw_tag_value(tag.name, tag.section_index)
            if not value or tag.name in cls.IGNORED_TAGS:
                continue
            scheme = urllib.parse.urlparse(value).scheme
            if (tag.name.startswith('Patch') or tag.name.startswith('Source')) and not scheme:
                # skip local sources
                continue

            # replace the whole version first
            updated_value = subversion_patterns[0][0].sub(subversion_patterns[0][1], value)
            # replace subversions only for remote sources/patches
            if tag.name.startswith('Patch') or tag.name.startswith('Source'):
                for sub_pattern, repl in subversion_patterns[1:]:
                    updated_value = sub_pattern.sub(repl, updated_value)
            rebase_spec_file.set_raw_tag_value(tag.name, updated_value, tag.section_index)

        for sec_index, (sec_name, section) in enumerate(rebase_spec_file.spec_content.sections):
            if sec_name.startswith('%changelog'):
                continue
            for index, line in enumerate(section):
                tag_ignored = any(MacroHelper.expand(line, line).startswith(tag) for tag in cls.IGNORED_TAGS)
                if index in examined_lines[sec_index] or tag_ignored:
                    continue
                start, end = spec_file.spec_content.get_comment_span(line, sec_name)
                updated_line = subversion_patterns[0][0].sub(subversion_patterns[0][1], line[:start])
                section[index] = updated_line + line[start:end]

        rebase_spec_file.save()



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import enum
import logging
import os
import shutil
from typing import Any, Dict, List, Optional, Type, Union, cast

from rebasehelper.logger import CustomLogger
from rebasehelper.plugins.plugin import Plugin
from rebasehelper.plugins.plugin_collection import PluginCollection
from rebasehelper.constants import RESULTS_DIR


logger: CustomLogger = cast(CustomLogger, logging.getLogger(__name__))


class CheckerCategory(enum.Enum):
    SOURCE: int = 1
    SRPM: int = 2
    RPM: int = 3


class BaseChecker(Plugin):
    """Base class of package checkers.

    Attributes:
        DEFAULT(bool): If True, the checker is run by default.
        CATEGORY(CheckerCategory): Category which determines when the checker is run.
        results_dir(str): Path where the results are stored.
    """

    DEFAULT: bool = False
    CATEGORY: Optional[CheckerCategory] = None
    results_dir: Optional[str] = None

    @classmethod
    def prepare_results_dir(cls):
        """Creates checker's results dir.

        Removes the existing results dir if it exists to avoid collisions.
        """
        if os.path.exists(cls.results_dir):
            shutil.rmtree(cls.results_dir)
        os.makedirs(cls.results_dir)

    @classmethod
    def get_checker_output_dir_short(cls):
        """Return short version of checker output directory"""
        return os.path.join(RESULTS_DIR, 'checkers', cls.name)

    @classmethod
    def is_available(cls):
        raise NotImplementedError()

    @classmethod
    def run_check(cls, results_dir, **kwargs):
        """Perform the check itself and return results."""
        raise NotImplementedError()

    @classmethod
    def format(cls, data):
        """Formats checker output to a readable text form."""
        raise NotImplementedError()

    @classmethod
    def get_category(cls):
        return cls.CATEGORY

    @classmethod
    def get_underlined_title(cls, text, separator='='):
        return "\n{}\n{}".format(text, separator * len(text))

    @classmethod
    def get_important_changes(cls, checker_output):  # pylint: disable=unused-argument
        """
        Each checker has an opportunity to highlight some of its output.
        This function is optional, as not all checkers provide output with critical information.

        Args:
            checker_output (dict): Dictionary with output from the given checker.

        Returns:
            list: List of strings to be output to CLI as warning messages.
        """
        return []


class CheckerCollection(PluginCollection):
    """
    Class representing the process of running various checkers on final packages.
    """

    def run(self, results_dir: str, checker_name: str, **kwargs: Any) -> Optional[Dict[str, Any]]:
        """Runs a particular checker and returns the results.

        Args:
            results_dir: Path to a directory in which the checker
                should store the results.
            checker_name: Name of the checker to be run.

        Raises:
            NotImplementedError: If a checker with the given name doesn't
                exist.

        Returns:
            Results of the checker.

        """
        try:
            checker = cast(Type[BaseChecker], self.plugins[checker_name])
        except KeyError:
            return None
        if checker.CATEGORY != kwargs.get('category'):
            return None
        if not checker.is_available():
            return None

        if checker.CATEGORY == CheckerCategory.SOURCE:
            logger.info("Running checks on source files using '%s'", checker_name)
        else:
            logger.info("Running checks on packages using '%s'", checker_name)
        return checker.run_check(results_dir, **kwargs)

    def get_supported_plugins(self) -> List[str]:
        return [k for k, v in self.plugins.items() if v and cast(Type[BaseChecker], v).is_available()]

    def get_default_plugins(self, return_one: bool = False) -> Union[str, List[str], None]:
        default = [k for k, v in self.plugins.items() if v and getattr(v, 'DEFAULT', False)]
        return default if not return_one else default[0] if default else None



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>


class RebaseHelperError(Exception):
    """Class representing Error raised inside rebase-helper after intentionally
    catching some expected and well known exception/error.

    """

    def __init__(self, *args, **kwargs):
        """Constructor of RebaseHelperError"""
        super(RebaseHelperError, self).__init__()
        if not args:
            self.msg = None
        elif len(args) > 1:
            self.msg = args[0] % args[1:]
        else:
            self.msg = args[0]
        self.logfiles = kwargs.get('logfiles')

    def __str__(self):
        return str(self.msg) if self.msg else ''


class CheckerNotFoundError(RuntimeError):
    """Error indicating not being able to find checker binary."""


class DownloadError(Exception):
    """Exception indicating that download of a file failed."""


class ParseError(Exception):
    pass


class LookasideCacheError(Exception):
    """Exception indicating a problem in accessing lookaside cache."""


class SourcePackageBuildError(RuntimeError):
    """Error indicating failure during the build of source package"""

    def __init__(self, *args, **kwargs):
        """Constructor of SourcePackageBuildError.

        Args:
            *args: List of arguments to be stored in the exception instance.
            **kwargs: Keyword arguments containing paths to logs with errors.

        """
        super(SourcePackageBuildError, self).__init__()
        self.args = args
        self.logfile = kwargs.get('logfile')
        self.logs = kwargs.get('logs')


class BinaryPackageBuildError(RuntimeError):
    """Error indicating failure during the build of binary package."""

    def __init__(self, *args, **kwargs):
        """Constructor of BinaryPackageBuildError.

        Args:
            *args: List of arguments to be stored in the exception instance.
            **kwargs: Keyword arguments containing paths to logs with errors.

        """
        super(BinaryPackageBuildError, self).__init__()
        self.args = args
        self.logfile = kwargs.get('logfile')
        self.logs = kwargs.get('logs')



import re

from argparse import ArgumentParser, SUPPRESS

from docutils import nodes
from docutils.statemachine import StringList
from docutils.parsers.rst import Directive
from docutils.parsers.rst.directives import flag, positive_int, unchanged, unchanged_required
from sphinx import addnodes
from sphinx.errors import SphinxError


class AutoArgsError(SphinxError):
    """Base class for AutoArgs errors"""
    category = 'autoargs extension error'


class AutoArgsDirective(Directive):
    """
    Generates program synopsis, description and option list from
    pre-configured ArgumentParser instance

    All sections and every option description can be extended with
    custom content

    :param function: function returning pre-configured ArgumentParser instance
    :param module: module containing the function
    :param program_name: name of the program
    :param synopsis_max_width: maximal width of synopsis (number of characters)
    :param ignore_option_groups: do not render options in groups
    """
    option_spec = dict(function=unchanged_required,
                       module=unchanged_required,
                       program_name=unchanged,
                       synopsis_max_width=positive_int,
                       ignore_option_groups=flag)
    has_content = True
    synopsis_max_width = None
    ignore_option_groups = False

    @staticmethod
    def _get_program_name(parser):
        """Return sanitized program name"""
        return re.sub(r'\s+', '-', parser.prog)

    def _decorate_references(self, text):
        """
        Decorate string conversion specifiers which are used as references
        to program name or action properties
        """
        # python string conversion specifier regex
        template = r'(%\({}\)[#0\- +]?(\d+|\*)?(\.(\d+|\*))?[hlL]?[diouxXeEfFgGcrs%])'
        replacements = dict(prog=r'\\ :program:`\1`\\ ',
                            default=r'\\ ``\1``\\ ',
                            metavar=r'\\ ``\1``\\ ')
        for k, v in replacements.items():
            text = re.sub(template.format(k), v, text or '')
        return text

    def _format_synopsis(self, synopsis):
        """
        Format synopsis so it fits into maximal width

        :param synopsis: synopsis text divided into non-breakable parts
        :return: list of lines of formatted synopsis text
        """
        def len_plain(text):
            text = re.sub(r':\w+:`(.+?)`', r'\1', text)
            text = re.sub(r'([`*]{1,2})(.+?)\1', r'\2', text)
            text = re.sub(r'\\ ', '', text)
            return len(text)
        result = []
        line = [synopsis.pop(0)]
        synopsis.reverse()
        while synopsis:
            line.append(synopsis.pop())
            if len_plain(' '.join(line)) > self.synopsis_max_width and len(line) > 2:
                synopsis.append(line.pop())
                result.append(' '.join(line))
                line = [' ']
        result.append(' '.join(line))
        return ['| {}'.format(l) for l in result]

    @staticmethod
    def _get_delimiter(fmt, action):
        """
        Extract delimiter from usage string formatted
        by specified HelpFormatter
        """
        if not action.option_strings or action.nargs == 0:
            return None
        else:
            usage = fmt._format_actions_usage([action], [])
            option_string = action.option_strings[0]
            idx = usage.find(option_string)
            if idx == -1:
                return None
            return usage[idx + len(option_string)]

    @classmethod
    def _get_option_group(cls, fmt, action):
        """
        Extract group of option properties from specified action

        :param fmt: HelpFormatter to use
        :param action: source Action
        :return: list of option properties
        """
        result = []
        if not action.option_strings:
            # positional option
            metavar, = fmt._metavar_formatter(action, action.dest)(1)
            args = fmt._format_args(action, action.dest)
            optional = False
            if args[0] == '[' and args[-1] == ']':
                args = args[1:-1]
                optional = True
            result.append(dict(name=metavar, synopsis=args, optional=optional))
        else:
            if action.nargs == 0:
                # option without arguments
                for option_string in action.option_strings:
                    result.append(dict(name=option_string, synopsis=option_string))
            else:
                # option with arguments
                args = fmt._format_args(action, action.dest.upper())
                delim = cls._get_delimiter(fmt, action)
                for option_string in action.option_strings:
                    result.append(dict(name=option_string, args=delim+args, synopsis=option_string+delim+args))
        return result

    def _build_option_description(self, fmt, action, custom_content):
        """
        Build description of program option

        :param fmt: HelpFormatter to use
        :param action: source Action
        :param custom_content: custom content for option
        :return: node forming option description
        """
        action.help = self._decorate_references(action.help)
        help = fmt._expand_help(action)
        result = nodes.container()
        self.state.nested_parse(StringList([help]), 0, result)
        if custom_content:
            if custom_content['action'] == 'append':
                result.extend(custom_content['content'].children)
            elif custom_content['action'] == 'prepend':
                result[0:0] = custom_content['content'].children
            elif custom_content['action'] == 'replace':
                result[:] = custom_content['content']
        return result

    def _build_option_index(self, progname, ids, names, synopses):
        """
        Build index for program option and register it in the environment

        :param progname: program name
        :param ids: list of all option ids
        :param names: list of all option names
        :param synopses: list of all option synopses
        :return: index node
        """
        env = self.state.document.settings.env
        result = addnodes.index(entries=[])
        for id, name, synopsis in zip(ids, names, synopses):
            env.domaindata['std']['progoptions'][progname, name] = env.docname, id
            if synopsis != name:
                env.domaindata['std']['progoptions'][progname, synopsis] = env.docname, id
            pair = '{} command line option; {}'.format(progname, ', '.join(synopses))
            result['entries'].append(('pair', pair, id, '', None))
        return result

    def _build_option(self, parser, action, custom_content):
        """
        Build single program option

        :param parser: pre-configured ArgumentParser instance
        :param action: source Action
        :param custom_content: custom content for options
        :return: node forming program option
        """
        def get_id(progname, name):
            id = 'cmdoption-{}'.format(progname)
            if not name.startswith('-'):
                id += '-arg-'
            id += name
            return id
        fmt = parser._get_formatter()
        result = nodes.container()
        signature = addnodes.desc_signature(ids=[], allnames=[], first=False)
        self.state.document.note_explicit_target(signature)
        description = self._build_option_description(fmt, action, custom_content)
        content = addnodes.desc_content('', description)
        result.append(addnodes.desc('', signature, content, objtype='option'))
        progname = self._get_program_name(parser)
        synopses = []
        for option in self._get_option_group(fmt, action):
            signature['ids'].append(get_id(progname, option['name']))
            signature['allnames'].append(option['name'])
            signature.append(addnodes.desc_name(text=option['name']))
            if 'args' in option:
                signature.append(addnodes.desc_addname(text=option['args']))
            signature.append(addnodes.desc_addname(text=', '))
            synopses.append(option['synopsis'])
        signature.pop()
        index = self._build_option_index(progname, signature['ids'], signature['allnames'], synopses)
        result.append(index)
        return result

    def _build_program_synopsis(self, parser):
        """
        Build program synopsis

        :param parser: pre-configured ArgumentParser instance
        :return: node forming program synopsis
        """
        fmt = parser._get_formatter()
        groups = []
        for action in parser._get_optional_actions() + parser._get_positional_actions():
            if action.help is SUPPRESS:
                continue
            in_group = False
            for group in parser._mutually_exclusive_groups:
                if action in group._group_actions:
                    in_group = True
                    if action == group._group_actions[0]:
                        groups.append(dict(actions=group._group_actions, required=group.required))
            if not in_group:
                groups.append(dict(actions=[action], required=action.required))
        synopsis = ['\\ :program:`{}`\\ '.format(parser.prog)]
        for group in groups:
            usages = []
            for action in group['actions']:
                option = self._get_option_group(fmt, action)[0]
                if option.get('optional') and len(group['actions']) == 1:
                    group['required'] = False
                usage = '\\ :option:`{}`\\ '.format(option['synopsis'])
                usages.append(usage)
            if not group['required']:
                synopsis.append('[{}]'.format(' | '.join(usages)))
            elif len(group['actions']) > 1:
                synopsis.append('({})'.format(' | '.join(usages)))
            else:
                synopsis.append('{}'.format(usages[0]))
        synopsis = self._format_synopsis(synopsis)
        paragraph = nodes.paragraph()
        self.state.nested_parse(StringList(synopsis), 0, paragraph)
        return nodes.container('', paragraph)

    def _build_program_description(self, parser):
        """
        Build program description

        :param parser: pre-configured ArgumentParser instance
        :return: node forming program description
        """
        description = self._decorate_references(parser.description)
        description = description % dict(prog=parser.prog)
        result = nodes.container()
        self.state.nested_parse(StringList([description]), 0, result)
        return result

    def _build_program_options(self, parser, custom_content):
        """
        Build list of program options

        :param parser: pre-configured ArgumentParser instance
        :param custom_content: custom content for options
        :return: node forming program options
        """
        result = nodes.container()
        if self.ignore_option_groups:
            actions = parser._get_positional_actions() + parser._get_optional_actions()
            actions = [a for a in actions if a.help is not SUPPRESS]
            for action in actions:
                cc = [v for k, v in custom_content.items() if k in action.option_strings]
                result.append(self._build_option(parser, action, cc[0] if cc else None))
        else:
            for group in parser._action_groups:
                actions = [a for a in group._group_actions if a.help is not SUPPRESS]
                if actions:
                    title = nodes.title(text=group.title.capitalize())
                    options = nodes.container()
                    for action in actions:
                        cc = [v for k, v in custom_content.items() if k in action.option_strings]
                        options.append(self._build_option(parser, action, cc[0] if cc else None))
                    result.append(nodes.section('', title, options, ids=[group.title.lower()]))
        return result

    def _get_custom_content(self):
        """Load custom content for each section and for program options"""
        content = nodes.container()
        self.state.nested_parse(self.content, self.content_offset, content)
        sections = {}
        options = {}
        if len(content) and isinstance(content[0], nodes.definition_list):
            for item in content[0]:
                term = None
                classifiers = []
                definition = None
                for element in item.children:
                    if isinstance(element, nodes.term):
                        if element.children:
                            term = str(element[0]).lower()
                    elif isinstance(element, nodes.classifier):
                        if element.children:
                            classifiers.append(str(element[0]).lower())
                    elif isinstance(element, nodes.definition):
                        if element.children:
                            definition = nodes.container('', *element.children)
                if term in ['synopsis', 'description', 'options', 'option']:
                    action = [c[1:] for c in classifiers if c and c[0] == '@']
                    action = action[0] if action else 'append'
                    opts = [c for c in classifiers if c and c[0] != '@']
                    if definition:
                        if term != 'option':
                            sections[term] = dict(action=action, content=definition)
                        else:
                            for opt in opts:
                                options[opt] = dict(action=action, content=definition)
        return sections, options

    def _construct_main_sections(self, parser):
        """
        Construct Synopsis, Description and Options sections

        :param parser: pre-configured ArgumentParser instance
        :return: list of section nodes
        """
        cc_sections, cc_options = self._get_custom_content()
        result = []
        for section in ['synopsis', 'description', 'options']:
            method = '_build_program_{}'.format(section)
            method = getattr(self, method)
            args = [parser]
            if section == 'options':
                args.append(cc_options)
            title = nodes.title(text=section.upper())
            content = method(*args)
            if section in cc_sections:
                cc = cc_sections[section]
                if cc['action'] == 'append':
                    content.extend(cc['content'].children)
                elif cc['action'] == 'prepend':
                    content[0:0] = cc['content'].children
                elif cc['action'] == 'replace':
                    content[:] = cc['content']
            else:
                # append empty paragraph to ensure separation from consecutive section
                content.append(nodes.paragraph(text=''))
            result.append(nodes.section('', title, content, ids=[section.lower()]))
        return result

    def _get_parser(self):
        """Get parser instance"""
        function = self.options.get('function')
        module = self.options.get('module')
        # import module and get instance of function to call
        parts = function.split('.')
        try:
            mod = __import__(module, globals(), locals(), [parts[0]])
        except ImportError as e:
            raise AutoArgsError('Problem importing module: {}'.format(str(e)))
        try:
            obj = getattr(mod, parts[0])
            for sub in parts[1:]:
                obj = getattr(obj, sub)
        except AttributeError as e:
            raise AutoArgsError('Problem accessing function: {}'.format(str(e)))
        # instantiate object
        try:
            parser = obj()
        except TypeError as e:
            raise AutoArgsError('Problem calling function: {}'.format(str(e)))
        if not isinstance(parser, ArgumentParser):
            raise AutoArgsError('Function must return instance of argparse.ArgumentParser or derived class')
        return parser

    def run(self):
        """Run the directive"""
        parser = self._get_parser()
        parser.prog = self.options.get('program_name', parser.prog)
        env = self.state.document.settings.env
        env.ref_context['std:program'] = self._get_program_name(parser)
        self.synopsis_max_width = self.options.get('synopsis_max_width', 80)
        self.ignore_option_groups = 'ignore_option_groups' in self.options
        env.autoargs_options = dict(synopsis_max_width=self.synopsis_max_width,
                                    ignore_option_groups=self.ignore_option_groups)
        return self._construct_main_sections(parser)


def setup(app):
    app.add_directive('autoargs', AutoArgsDirective)
    return dict(version='0.1')



# -*- coding: utf-8 -*-
#
# This tool helps you rebase your package to the latest version
# Copyright (C) 2013-2019 Red Hat, Inc.
#
# This program is free software; you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation; either version 2 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License along
# with this program; if not, write to the Free Software Foundation, Inc.,
# 51 Franklin Street, Fifth Floor, Boston, MA 02110-1301 USA.
#
# Authors: Petr Hráček <phracek@redhat.com>
#          Tomáš Hozza <thozza@redhat.com>
#          Nikola Forró <nforro@redhat.com>
#          František Nečas <fifinecas@seznam.cz>

import locale


PROGRAM_DESCRIPTION: str = 'Tool to help package maintainers rebase their packages to the latest upstream version'
NEW_ISSUE_LINK: str = 'https://github.com/rebase-helper/rebase-helper/issues/new'

RESULTS_DIR: str = 'rebase-helper-results'
WORKSPACE_DIR: str = 'rebase-helper-workspace'

REBASED_SOURCES_DIR: str = 'rebased-sources'
OLD_BUILD_DIR: str = 'old-build'
NEW_BUILD_DIR: str = 'new-build'
CHECKERS_DIR: str = 'checkers'

LOGS_DIR: str = 'logs'
DEBUG_LOG: str = 'debug.log'
TRACEBACK_LOG: str = 'traceback.log'
VERBOSE_LOG: str = 'verbose.log'
INFO_LOG: str = 'info.log'
REPORT: str = 'report'
CHANGES_PATCH: str = 'changes.patch'

OLD_SOURCES_DIR: str = 'old_sources'
NEW_SOURCES_DIR: str = 'new_sources'

GIT_CONFIG: str = '.gitconfig'

CONFIG_PATH: str = '$XDG_CONFIG_HOME'
CONFIG_FILENAME: str = 'rebase-helper.cfg'

SYSTEM_ENCODING: str = locale.getpreferredencoding()
